############################################################################################################
# VaRank 1.3.2                                                                                             #
#                                                                                                          #
# VaRank: a simple and powerful tool for ranking genetic variants                                          #
#                                                                                                          #
# Copyright (C) 2015 Veronique Geoffroy (veronique.geoffroy@inserm.fr)                                     #
#                    Jean Muller (jeanmuller@unistra.fr)                                                   #
#                                                                                                          #
# Please cite the following article:                                                                       #
# Geoffroy V.*, Pizot C.*, Redin C., Piton A., Vasli N., Stoetzel C., Blavier A., Laporte J. and Muller J. #
# VaRank: a simple and powerful tool for ranking genetic variants.                                         #
# PeerJ. 2015 (10.7717/peerj.796)                                                                          #
#                                                                                                          #
# This is part of VaRank source code.                                                                      #
#                                                                                                          #
# This program is free software; you can redistribute it and/or                                            #
# modify it under the terms of the GNU General Public License                                              #
# as published by the Free Software Foundation; either version 3                                           #
# of the License, or (at your option) any later version.                                                   #
#                                                                                                          #
# This program is distributed in the hope that it will be useful,                                          #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                                           #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                            #
# GNU General Public License for more details.                                                             #
#                                                                                                          #
# You should have received a copy of the GNU General Public License                                        #
# along with this program; If not, see <http://www.gnu.org/licenses/>.                                     #
############################################################################################################


## Return the bonus value to add to the VaRank score
proc addBonus {siftPred siftMed pph2 phastcons} {

    global g_VaRank

    if {[info exists g_VaRank(DEBUG)]} {puts "Add Bonus values: $siftPred $siftMed $pph2 $phastcons"}
    
    set bonus 0
    
    if {[info exists g_VaRank(snpeffDir)]} {
	if {[regexp -nocase "deleterious" $siftPred]} {
	    incr bonus $g_VaRank(B_SIFT)
	}
	if {[regexp "probably damaging|deleterious" $pph2]} {
	    incr bonus $g_VaRank(B_PPH2)
	}	
        if {[regexp "probablydamaging|deleterious" $pph2]} {
            incr bonus $g_VaRank(B_PPH2)
        }
    } else {
	if {[string is double $siftMed]} {
	    if {[regexp "Deleterious" $siftPred] && $siftMed >= 2.75 && $siftMed <= 3.5} {
		incr bonus $g_VaRank(B_SIFT)
	    }
	}
	if {[regexp "deleterious" $pph2]} {
	    incr bonus $g_VaRank(B_PPH2)
	}
    }
    if {[string is double $phastcons] && $phastcons > 0.95} {incr bonus $g_VaRank(B_phastCons)}
    
    if {[info exists g_VaRank(DEBUG)]} {puts "Bonus $bonus"}
    
    return $bonus
}

proc scoreAllTheID {} {

    global g_VaRank
    global g_ANNOTATION
    global g_lScore
    global g_deltaSSF
    global g_deltaMES
    global g_deltaNNS

    ## Define the "g_lScore" global variable with the genetic variants analysed 
    ###########################################################################
    puts "...scoring each genetic variant ([clock format [clock seconds] -format "%B %d %Y - %H:%M"])"
    regsub -all "^\\\\?{|\\\\?}$" [split $g_ANNOTATION(#id) "\t"] "" L
    set i_gene      [lsearch -exact $L "gene"];           if {$i_gene   == -1} {puts "column number not found for gene - Exit"; exit}
    
    set i_trans     [lsearch -exact $L "transcript"];     if {$i_trans  == -1} {puts "column number not found for transcript - Exit"; exit}
    
    set i_length    [lsearch -exact $L "transLen"];       if {$i_length == -1} {puts "column number not found for transLen - Exit"; exit}
    set i_clinic    [lsearch -exact $L "rsClinicalSignificance"]; if {$i_clinic == -1} {puts "column number not found for rsClinicalSignificance - Exit"; exit}
    set i_clinic2   [lsearch -exact $L "clinVarClinSignifs"]; if {$i_clinic2 == -1} {puts "column number not found for clinVarClinSignifs - Exit"; exit}
    set i_effect    [lsearch -exact $L "codingEffect"];   if {$i_effect == -1} {puts "column number not found for codingEffect - Exit"; exit}

    set i_distSS    [lsearch -exact $L "distNearestSS"];  if {$i_distSS == -1} {puts "column number not found for distNearestSS - Exit"; exit}
    set i_nearSS    [lsearch -exact $L "nearestSSType"];  if {$i_nearSS == -1} {puts "column number not found for nearestSSType - Exit"; exit}
    set i_lse       [lsearch -exact $L "localSpliceEffect"];  if {$i_lse == -1} {puts "column number not found for localSpliceEffect - Exit"; exit}
    set i_intron    [lsearch -exact $L "varLocation"];    if {$i_intron == -1} {puts "column number not found for varLocation - Exit"; exit}
    set i_wtSSF     [lsearch -exact $L "wtSSFScore"];     if {$i_wtSSF  == -1} {puts "column number not found for wtSSFScore - Exit"; exit}
    set i_wtMES     [lsearch -exact $L "wtMaxEntScore"];  if {$i_wtMES  == -1} {puts "column number not found for wtMaxEntScore - Exit"; exit}
    set i_wtNNS     [lsearch -exact $L "wtNNSScore"];     if {$i_wtNNS  == -1} {puts "column number not found for wtNNSScore - Exit"; exit}
    set i_varSSF    [lsearch -exact $L "varSSFScore"];    if {$i_varSSF == -1} {puts "column number not found for varSSFScore - Exit"; exit}
    set i_varMES    [lsearch -exact $L "varMaxEntScore"]; if {$i_varMES == -1} {puts "column number not found for varMaxEntScore - Exit"; exit}
    set i_varNNS    [lsearch -exact $L "varNNSScore"];    if {$i_varNNS == -1} {puts "column number not found for varNNSScore - Exit"; exit}

    set i_siftPred  [lsearch -exact $L "SIFTprediction"]; if {$i_siftPred  == -1} {puts "column number not found for SIFTprediction - Exit"; exit}
    set i_siftMed   [lsearch -exact $L "SIFTmedian"];     if {$i_siftMed   == -1} {puts "column number not found for SIFTmedian - Exit"; exit}
    set i_pph2      [lsearch -exact $L "PPH2class"];      # if $PPH doesn't exists, $i_pph2 == -1
    set i_phastcons [lsearch -exact $L "phastCons"];      if {$i_phastcons == -1} {puts "column number not found for phastCons - Exit"; exit}
    set i_hgmdId    [lsearch -exact $L "hgmdId"];         if {$i_hgmdId    == -1} {puts "column number not found for hgmdId - Exit"; exit}
    #set i_hgmdPhen  [lsearch -exact $L "hgmdPhenotype"];  if {$i_hgmdPhen  == -1} {puts "column number not found for hgmdPhenotype - Exit"; exit}
    #set i_hgmdPub   [lsearch -exact $L "hgmdPubMedId"];   if {$i_hgmdPub   == -1} {puts "column number not found for hgmdPubMedId - Exit"; exit}

    #set i_phylop [lsearch -exact $L "phyloP"]; if {$i_phylop == -1} {puts "column number not found for phyloP - Exit"; exit}
    
    foreach ID [array name g_ANNOTATION] {
	if {$ID == "#id"} {continue}
	set score      0
	set bestScore  0
	set bestLength 0
	set allGene   {}
	set bestGene ".."

	if {[info exists g_VaRank(DEBUG)]} {puts "Variant: $ID"}

	#For a single variant, multiple isoforms can give different score. Each is evaluated.
	foreach L $g_ANNOTATION($ID) {
	    set L [split $L "\t"]

	    #Reset score between isoforms
	    set score 0

	    set trans     ""
	    set length    ""
	    set gene      ""
	    set effect    ""
	    set clinic    ""
	    set hgmdId    ""
	    set siftPred  ""
	    set siftMed   ""
	    set pph2      ""
	    set phastcons ""


	    if {[info exists g_VaRank(DEBUG)]} {puts "Data: $L"}
	    
	    set trans [lindex $L $i_trans]  

	    set length [lindex $L $i_length]   
	    set gene [lindex $L $i_gene]     
	    if {![regexp -nocase "$gene" $allGene]} {lappend allGene $gene}
	    
	    set effect [lindex $L $i_effect]   
	    set clinic [lindex $L $i_clinic]   
	    append clinic " [lindex $L $i_clinic2]"

	    set hgmdId [lindex $L $i_hgmdId]   
	    
	    set siftPred [lindex $L $i_siftPred] 
	    set siftMed [lindex $L $i_siftMed]  
	    if {$i_pph2!=-1} {set pph2 [lindex $L $i_pph2]}
	    set phastcons [lindex $L $i_phastcons]

	    #puts "effect $effect, pph2 $pph2"

	    #General scoring scheme for big categories
	    #
	    regsub -all "non-pathogenic" $clinic "" clinicChanged
	    if {[regexp -nocase "pathogenic" $clinicChanged] || ($hgmdId != "" && [regexp -nocase "^CM" $hgmdId])} {
		set score $g_VaRank(S_Known)
		#incr score [addBonus $siftPred $siftMed $pph2 $phastcons]
	    } elseif {[regexp "nonsense|stop_gained" $effect]} {
		set score $g_VaRank(S_Nonsense)
		#incr score [addBonus $siftPred $siftMed $pph2 $phastcons]
		incr score [addBonus "NA" "NA" "NA" $phastcons]
	    } elseif {[regexp -nocase "Frameshift" $effect]} {
		set score $g_VaRank(S_Fs)
		#incr score [addBonus $siftPred $siftMed $pph2 $phastcons]
	    } elseif {[regexp "startloss|start_lost|start loss" $effect]} {
		set score $g_VaRank(S_StartLoss)
		#incr score [addBonus $siftPred $siftMed $pph2 $phastcons]
		incr score [addBonus "NA" "NA" "NA" $phastcons]
	    } elseif {[regexp "stoploss|stop_lost|stop loss" $effect]} {
		set score $g_VaRank(S_StopLoss)
		#incr score [addBonus $siftPred $siftMed $pph2 $phastcons]
		incr score [addBonus "NA" "NA" "NA" $phastcons]
	    } elseif {[regexp "missense|rare_amino_acid" $effect]} {
		set score $g_VaRank(S_Missense)
		incr score [addBonus $siftPred $siftMed $pph2 $phastcons]
	    } elseif {[regexp "In-frame|inframe" $effect]} {
		set score $g_VaRank(S_Inframe)
		#incr score [addBonus $siftPred $siftMed $pph2 $phastcons]
	    } elseif {[regexp "synonymous|start_retained|stop_retained" $effect]} {
		set score $g_VaRank(S_Synonymous)
		#incr score [addBonus $siftPred $siftMed $pph2 $phastcons]
		incr score [addBonus "NA" "NA" "NA" $phastcons]
	    }

	    if {[info exists g_VaRank(DEBUG)]} {puts "Score $score"}

	    #Scoring for splice effect
	    #
	    set distSS [lindex $L $i_distSS]
	    set nearSS [lindex $L $i_nearSS] 
	    set lse [lindex $L $i_lse]
	    set intron [lindex $L $i_intron]
	    set wtSSF [lindex $L $i_wtSSF] 
	    set wtMES [lindex $L $i_wtMES] 
	    set wtNNS [lindex $L $i_wtNNS] 
	    set varSSF [lindex $L $i_varSSF]
	    set varMES [lindex $L $i_varMES] 
	    set varNNS [lindex $L $i_varNNS] 
	    
	    set n   0
	    set tot 0
	    
	    #2013/03/15 bug in alamut when predictions for Variant is null the score is empty and not equal to  0
	    #2013/03/26 bug splice scores are not indexed by isoforms/transcript
	    #
	    if {[info exists g_VaRank(DEBUG)]} {puts "Splices Scores: $wtSSF, $wtMES, $wtNNS, $varSSF, $varMES, $varNNS"}
	    

	    if {$wtSSF != "" && $wtSSF != "NA" && $wtSSF != "0"} {
		if {$varSSF=="" || $varSSF=="NA"} {set varSSF 0}
		set  g_deltaSSF($ID,$trans) [format "%.1f" [expr {($varSSF-$wtSSF)/$wtSSF*100.0}]]
		if {$g_deltaSSF($ID,$trans) < $g_VaRank(SSFcutoff)} {incr n}
		incr tot
	    } else {set g_deltaSSF($ID,$trans) "NA"}

	    if {[info exists g_VaRank(DEBUG)]} {puts "g_deltaSSF $trans [set g_deltaSSF($ID,$trans)]"}

	    if {$wtMES != "" && $wtMES != "NA" && $wtMES != "0"} {
		if {$varMES == "" || $varMES=="NA"} {set varMES 0}

		set  g_deltaMES($ID,$trans) [format "%.1f" [expr {($varMES-$wtMES)/$wtMES*100.0}]]
		if {$g_deltaMES($ID,$trans) < $g_VaRank(MEScutoff)} {incr n}
		incr tot
	    } else {set g_deltaMES($ID,$trans) "NA"}

	    if {[info exists g_VaRank(DEBUG)]} {puts "g_deltaMES $trans [set g_deltaMES($ID,$trans)]"}

	    if {$wtNNS != "" && $wtNNS != "NA"  && $wtNNS != "0"} {
		if {$varNNS == "" || $varNNS=="NA"} {set varNNS 0}

		set  g_deltaNNS($ID,$trans) [format "%.1f" [expr {($varNNS-$wtNNS)/$wtNNS*100.0}]]
		if {$g_deltaNNS($ID,$trans) < $g_VaRank(NNScutoff)} {incr n}
		incr tot
	    } else {set g_deltaNNS($ID,$trans) "NA"}

	    if {[info exists g_VaRank(DEBUG)]} {puts "g_deltaNNS $trans [set g_deltaNNS($ID,$trans)]"}

	    set splicingEffect 0

	    #JEAN Correct when only one program available
	    #
	    if {$tot != 0 && [expr {$n*1.0/$tot}] >= 0.5} {set splicingEffect 1}

	    if {[info exists g_VaRank(DEBUG)]} {puts "splicingEffect $splicingEffect"}

	    if {$splicingEffect} {
		if {$score <= $g_VaRank(S_EssentialSplice)} {
		    # Essential splice site
		    if {$nearSS == "5'" && [regexp "^1$|^2$" $distSS]} {
			set  score $g_VaRank(S_EssentialSplice)
			#incr score [addBonus $siftPred $siftMed $pph2 $phastcons]
			incr score [addBonus "NA" "NA" "NA" $phastcons]
		    } elseif {$nearSS == "3'" && [regexp "^-1$|^-2$" $distSS]} {
			set  score $g_VaRank(S_EssentialSplice)
			#incr score [addBonus $siftPred $siftMed $pph2 $phastcons]
			incr score [addBonus "NA" "NA" "NA" $phastcons]
		    }
		} 
		if {$score <= $g_VaRank(S_CloseSplice)} {
		    # Intron-Exon boundary
		    if {$nearSS == "5'" && $distSS >= -3 && $distSS <= 6} {
			set  score $g_VaRank(S_CloseSplice)
			#incr score [addBonus $siftPred $siftMed $pph2 $phastcons]
			incr score [addBonus "NA" "NA" "NA" $phastcons]
		    } elseif {$nearSS == "3'" && $distSS >= -12 && $distSS <= 2} {
			set  score $g_VaRank(S_CloseSplice)
			#incr score [addBonus $siftPred $siftMed $pph2 $phastcons]
			incr score [addBonus "NA" "NA" "NA" $phastcons]
		    }
		} 
		if {$score <= $g_VaRank(S_DeepSplice)} {
		    # Deep intron-exon boundary
		    if {$intron == "intron"} {
			set score $g_VaRank(S_DeepSplice)
			#incr score [addBonus $siftPred $siftMed $pph2 $phastcons]
			incr score [addBonus "NA" "NA" "NA" $phastcons]
		    }
		}
	    }
	    ## Use of "LocalSpliceEffect" field from Alamut (even if $splicingEffect == 0):
	    ## VaRank score = 40 : New Donor Site, New Acceptor Site, Cryptic Donor Strongly Activated, Cryptic Acceptor Strongly Activated,
	    ## VaRank score = 35 : Cryptic Donor Weakly Activated, Cryptic Acceptor Weakly Activated
	    if {[regexp "New|Strongly" $lse] && $score<$g_VaRank(S_LSEstrong)} {set score $g_VaRank(S_LSEstrong)}
	    if {[regexp "Weakly" $lse] && $score<$g_VaRank(S_LSEweak)} {set score $g_VaRank(S_LSEweak)}

	    
	    if {$score == 0} {incr score [addBonus $siftPred $siftMed $pph2 $phastcons]}

	    if {[info exists g_VaRank(DEBUG)]} {puts "Score $ID $trans: $score / Best $bestScore"}
	    
	    #Within the different isoforms we keep the one for which the effect is maximum
	    #If similar score between isoforms we keep the longest one
	    if {$score > $bestScore} {
		set bestGene $gene
		set bestL $L
		set bestScore $score
		set bestLength $length
	    } elseif {$score == $bestScore} {
		if {$length > $bestLength} {
		    set bestGene $gene
		    set bestL $L
		    set bestLength $length
		}
	    }
	}
	if {$bestGene != $allGene} {
	    regsub "$bestGene" $allGene "" allGene
	    set allGene "$bestGene/[join $allGene "/"]"
	}

	set g_ANNOTATION($ID) [join [lreplace $bestL $i_gene $i_gene "$allGene"] "\t"]
	
	if {[info exists g_VaRank(DEBUG)]} {puts "Best $ID lreplace $bestL $i_gene $i_gene $allGene"}
	
	lappend g_lScore "$ID $bestScore"

    }

    puts "...classifying each genetic variant ([clock format [clock seconds] -format "%B %d %Y - %H:%M"])"

    if {![info exists g_lScore] || $g_lScore == {}} {
	puts "No score could be retrieved something must be wrong with variation identifiers (g_lScore) - Exit";exit
	if {[info exists g_VaRank(DEBUG)]} {
	    puts "List of uploaded variation from alamut: [array name g_ANNOTATION]"
	    puts "g_lScore $g_lScore"
	}
    }
    
    set g_lScore [lsort -command DescendingSortOnElement1 $g_lScore]

    return
}



